isr:
  cli
  push ax
  cmp ah, 0 ;function 0 - get local information
  jne not_0
  jmp get_local_info
not_0:
  cmp ah, 1 ;function 1 - disable and uninstall MiniTCP
  jne not_1
  jmp uninstall
not_1:
  cmp ah, 2 ;function 2 - send UDP packet
  jne not_2
  jmp send_udp2
not_2:
  cmp ah, 3 ;function 3 - define callback pointer for UDP receive
  jne not_3
  jmp define_callback_udp
not_3:
  cmp ah, 5 ;function 5 - define callback pointer for TCP
  jne not_5
  jmp define_callback_tcp
not_5:
  cmp ah, 6 ;function 6 - establish outgoing TCP connection
  jne not_6
  jmp goto_tcp_connect2
not_6:
  cmp ah, 7 ;function 7 - establish a TCP listener socket
  jne not_7
  jmp goto_tcp_listen
not_7:
  cmp ah, 8 ;function 8 - send TCP data packet
  jne not_8
  jmp send_tcp_new_pkt2
not_8:
  cmp ah, 9 ;function 9 - close TCP socket/handle
  jne not_9
  jmp tcp_close_handle
not_9:
  cmp ah, 10 ;function 10 - resume receive on a handle
  je tcp_resume_recv2

isr_finish:
  pop ax
  iret

get_local_info:
  mov ax, cs
  mov ds, ax
  mov si, offset local_ip
  pop ax
  iret

save_new_handle dw 65535
goto_tcp_listen:
  pushall
  mov dx, cs:[tcp_cur_handle]
  mov cs:[save_old_handle4], dx
  mov cs:[tcp_cur_handle], bx

  mov cs:[make_listener], 1
  call tcp_connect
  mov cs:[save_new_handle], bx
  mov cs:[make_listener], 0

  mov dx, cs:[save_old_handle4]
  mov cs:[tcp_cur_handle], dx
  popall
  pop ax
  mov bx, cs:[save_new_handle]
  iret

define_callback_tcp:
  mov cs:[tcp_callback_seg], bx
  mov cs:[tcp_callback_ptr], cx
  pop ax
  iret

send_udp2:
  jmp send_udp

tcp_resume_recv2:
  jmp tcp_resume_recv

uninstall:
  call release_type
  mov ax, cs
  mov es, ax
  mov ah, 49h ;DOS 2.0+ - free memory
  int 21h
  jmp isr_finish

send_tcp_new_pkt2:
  jmp send_tcp_new_pkt

define_callback_udp:
  mov cs:[udp_callback_seg], bx
  mov cs:[udp_callback_off], cx
  pop ax
  iret

goto_tcp_connect2:
  jmp goto_tcp_connect


tcp_close_handle:
  pushall
  mov dx, cs:[tcp_cur_handle]
  mov cs:[save_old_handle4], dx
  mov cs:[tcp_cur_handle], bx

  cmp al, 1
  je abort_immediately

  call tcp_init_normal_close
  mov dx, cs:[save_old_handle4]
  mov cs:[tcp_cur_handle], dx
  popall
  pop ax
  iret

  abort_immediately:
  call tcp_reset

  mov dx, cs:[save_old_handle4]
  mov cs:[tcp_cur_handle], dx
  popall
  pop ax
  iret

tcp_resume_recv:
  pushall
  mov dx, cs:[tcp_cur_handle]
  mov cs:[save_old_handle4], dx
  mov cs:[tcp_cur_handle], bx

  call tcp_send_empty_ack
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  mov si, di
  add di, offset tcp_ack
  add si, offset tcp_last_recv
  mov ax, cs:[di]
  mov cs:[si], ax
  mov ax, cs:[di+2]
  mov cs:[si+2], ax

  mov dx, cs:[save_old_handle4]
  mov cs:[tcp_cur_handle], dx
  popall
  pop ax
  iret

save_old_handle4 dw 0
send_tcp_new_pkt:
  pushall
  mov dx, cs:[tcp_cur_handle]
  mov cs:[save_old_handle4], dx
  mov cs:[tcp_cur_handle], bx

  mov di, offset tcp_state
  cmp byte ptr cs:[bx+di], TCP_STATE_ESTABLISHED
  jne send_tcp_new_fail

  ;mov di, offset tcp_ack_wait
  ;cmp byte ptr cs:[bx+di], 1
  ;je send_tcp_new_fail

  mov ax, ds
  mov di, bx
  shl di, 1
  add di, offset tcp_send_seg
  mov cs:[di], ax
  mov di, bx
  shl di, 1
  add di, offset tcp_send_ptr
  mov cs:[di], si
  mov di, bx
  shl di, 1
  add di, offset tcp_send_len
  mov cs:[di], cx
  push cx
  push bx
  call tcp_data_out
  pop bx
  pop cx

  mov di, bx
  shl di, 1
  shl di, 1
  add di, offset tcp_seq
  mov ax, word ptr cs:[di+2]
  mov dx, word ptr cs:[di]
  xchg ah, al
  xchg dh, dl
  add ax, cx
  adc dx, 0
  xchg ah, al
  xchg dh, dl
  mov word ptr cs:[di+2], ax
  mov word ptr cs:[di], dx
  ;;dec cx
  ;add_send_len_loop: ;this could be done WAY faster... fix it, you lazy bum!
  ;	add byte ptr cs:[di+3], 1
  ;	adc byte ptr cs:[di+2], 0
  ;	adc byte ptr cs:[di+1], 0
  ;	adc byte ptr cs:[di], 0
  ;	loop add_send_len_loop
  mov di, offset tcp_ack_wait
  mov byte ptr cs:[bx+di], 0

  mov dx, cs:[save_old_handle4]
  mov cs:[tcp_cur_handle], dx
  popall
  pop ax
  xor al, al ;AL = 0 on successful return from TCP send
  iret
send_tcp_new_fail:
  mov dx, cs:[save_old_handle4]
  mov cs:[tcp_cur_handle], dx
  popall
  pop ax
  mov al, 1 ;AL = 1 on failure return from TCP send
  iret

send_udp:
  cld
  pushall
  push bx
  mov bx, es
  xchg bh, bl ;byteswap source port
  mov word ptr cs:[ip_send_data], bx ;source port 0
  pop bx
  mov ax, cs
  mov es, ax
  xchg bh, bl ;byteswap destination port
  mov cs:[ip_send_data+2], bx
  add cx, 8 ;add header size to length
  xchg ch, cl ;byteswap length
  mov cs:[ip_send_data+4], cx
  xchg ch, cl ;change it back for us to use again
  sub cx, 8 ;subtract header size
  mov cs:[ip_send_data+6], 0 ;null checksum, we're not worrying about it for now

  push es
  push ds
  push di
  push si
  push cx
  mov cx, cs
  mov es, cx
  mov ds, dx
  mov cx, 2
  mov si, di
  mov di, offset ip_send_dest
  rep movsw
  pop cx
  pop si
  pop di
  pop ds
  pop es

  push cx
  mov di, offset ip_send_data+8
  rep movsb ;copy the data to our outgoing buffer
  pop cx
  add cx, 8 ;add 8 bytes to the total length for our UDP header
  mov cs:[ip_send_total_len], cx ;and save it in the IP send packet header

  mov ax, 0FFFFh
  mov cx, 3
  mov di, offset pkt_dest
  rep stosw

  mov cs:[ip_send_protocol], 17 ;17 is UDP
  call ip_send

  ;push ds
  ;push ax
  ;mov ax, cs
  ;mov ds, ax
  ;mov si, offset strsentudp
  ;call printmsg
  ;pop ax
  ;pop ds

  popall
  pop ax
  iret

temp_flags dw 0
temp_dx dw 0
goto_tcp_connect:
  call tcp_connect
  pop ax
  iret

strintcall db '[DEBUG] Interrupt called!', 13, 10, 0
strsentudp db '[DEBUG] Sent UDP packet!', 13, 10, 0
strdefcb   db '[DEBUG] Defined callback pointer!', 13, 10, 0
