;MiniTCP TSR - a TCP/IP protocol stack for DOS
;(c)2011 Mike Chambers
;note: TCP support was never finished. it works but can be buggy.
;      the UDP support appears to work properly. i stopped work
;      on this TSR to favor of writing a new TCP stack as a C library.
  
.model tiny
.code

include macros.inc
include defs.inc

org 100h

entry_point:
jmp main ;skip over some variables

local_ip    db 192, 168,   1,   87
subnet      db 255, 255, 255,   0
gateway_ip  db 192, 168,   1,   1
local_mac   db 6 dup (0)
gateway_mac db 6 dup (0)
tcp_intr    db 61h

mem_ceiling dw ?

main:
  push cs
  pop ds
  push cs
  pop es
  push cs
  pop ss
  ; give us 1 KB for a stack after the packet buffers
  mov sp, offset recv_bufs
  ;add sp, (MTU+14)*INCOMING_BUFS
  ;add sp, 1024
  mov sp, 32766
  mov cs:[mem_ceiling], sp
  sub sp, 2

  mov si, offset strwelcome
  call printmsg
  mov si, offset strinit
  call printmsg

  call access_type
  call get_address

  mov si, offset strokay
  call printmsg
  mov si, offset strhandle
  call printmsg
  mov ax, cs:[handle]
  call print_dec
  mov si, offset strhandle2
  call printmsg
  mov si, offset strourmac
  call printmsg
  mov si, offset local_mac
  call printmac
  call crlf
  mov si, offset strourip
  call printmsg
  mov si, offset local_ip
  call printip
  call crlf
  call set_interrupts
  call resolve_gateway_mac

  jmp go_tsr


;the stuff below is useful while debugging internal stuff, if jmp go_tsr
;is commented out, we'll just run this code and not go TSR.
  sti
  waitforkey:
	mov ah, 1
	int 16h
	jnz waitforkey

  ;now remove that key from the buffer, because we're thoughtful like that.
  mov ah, 0
  int 16h

  ;release_type to give our handle back to the packet driver
  call release_type

  mov si, offset strrecvcount
  call printmsg
  mov ax, cs:[recv_count]
  call print_dec
  call crlf

  mov si, offset strdropcount
  call printmsg
  mov ax, cs:[drop_count]
  call print_dec
  call crlf

  jmp good_exit

  ;ret

set_interrupts:
  mov ax, cs
  mov ds, ax
  mov ax, 2561h ;DOS 1.0+ - set interrupt vector
  mov dx, offset isr
  int 21h

  mov ax, 3508h ;DOS 1.0+ - get interrupt vector
  int 21h
  mov ax, es
  mov cs:[old_timer_seg], ax
  mov cs:[old_timer_ptr], bx

  mov ax, cs
  mov ds, ax
  mov ax, 2508h ;DOS 1.0+ - set interrupt vector
  mov dx, offset timer_isr
  int 21h
  ret

go_tsr:
  sti
  ;lets go TSR
  mov ax, 3100h ;DOS 2.0+ - terminate and stay resident (AL = return code)
  mov dx, 2048 ;32 KB ;1152 ;keep 18 KB of RAM (count in DX is 16-byte paragraphs)
  int 21h

good_exit:
  mov ax, 4C00h ;DOS 2.0+ - terminate with return code (AL = 0 for normal)
  int 21h

error_exit: ;whatever failed puts an error code into CX before calling this
  push cs
  pop ds ;just to be sure
  mov si, offset strerror ;point DS:SI to "Fatal error: " string
  call printmsg ;display it

  mov ax, 55 ;the error strings are 55 bytes each
  mul cx ;multiply 55 with err num
  mov si, offset err_strings ;point DS:SI to base of error strings array
  add si, ax ;and add the product from 55*errnum for our specific err string
  call printmsg ;display it
  call crlf ;and display a CR-LF pair

  mov ax, 4C01h ;DOS 2.0+ - terminate with return code (AL = error code)
  int 21h ;sayonara, mi amigos!

resolve_gateway_mac:
  pushf
  mov si, offset strgateway
  call printmsg
  cld
  ;lets form the ARP request packet
  mov ax, cs
  mov ds, ax
  mov es, ax
  mov di, offset pkt_data
  mov ax, 0100h ;1 for hardware type, 16-bit network-byte order
  stosw
  mov ax, 0008h ;0x0800 for protocol type (IP)
  stosw
  mov al, 6 ;hardware address length
  stosb
  mov al, 4 ;protocol address length
  stosb
  mov ax, 0100h ;1 for request operation packet
  stosw
  mov si, offset local_mac
  mov cx, 3
  rep movsw
  mov si, offset local_ip
  mov cx, 2
  rep movsw
  xor ax, ax ;null for target hardware address, as we obviously don't know it
  mov cx, 3
  rep stosw
  mov si, offset gateway_ip
  mov cx, 2
  rep movsw
  ;ARP packet is ready, fill in ethernet packet fields
  mov di, offset pkt_dest
  mov ax, 0FFFFh
  mov cx, 3
  rep stosw
  mov si, offset local_mac
  mov cx, 3
  rep movsw
  mov ax, 0608h ;0x0806 ethertype in network-byte order for ethertype
  stosw

  mov cx, 28 ;set packet size to 28 bytes
  mov word ptr cs:[timers], 0 ;clear first timer for our timeout checking
  mov word ptr cs:[timers+2], 0 ;use second for resend scheduling
  call send_pkt ;send it!

  ;wait for ARP reply, 5 seconds is good. rebroadcast ARP req every 1 sec.
  sti
  gw_reply_wait:
	mov si, offset gateway_mac
	mov cx, 3
	gw_mac_check:
	  lodsw
	  cmp ax, 0
	  jnz gw_finished ;if something in the gateway MAC isn't 0, we've got it!
	  loop gw_mac_check
	cmp word ptr cs:[timers], 3000 ;5 seconds = 5000 milliseconds
	jae gw_failed
	cmp word ptr cs:[timers+2], 1000 ;rebroadcast request every second
	jb gw_reply_wait
	mov word ptr cs:[timers+2], 0
	mov cx, 28 ;rebroadcast 28-byte packet
	call send_pkt
	jmp gw_reply_wait

  gw_finished:
  mov si, offset gateway_ip
  call printip
  mov si, offset strisat
  call printmsg
  mov si, offset gateway_mac
  call printmac
  call crlf
  popf
  ret

  gw_failed:
  mov si, offset strgwfail
  call printmsg
  popf
  ret

TIMER_COUNT equ 20
timers dw TIMER_COUNT dup (0) ;16-bit millisecond timers array
timer_isr:
  cli
  pushf
  cld
  push cx
  push ds
  push si
  mov cx, cs
  mov ds, cx
  mov si, offset timers
  mov cx, TIMER_COUNT ;update all timers
  timer_update:
	add word ptr [si], 55 ;55 milliseconds per timer tick. close enough, anyway.
	add si, 2
	loop timer_update
  inc cs:[tcp_srcport]
  call tcp_process_tick
  pop si
  pop ds
  pop cx
  popf
  db 0EAh
  old_timer_ptr dw 0
  old_timer_seg dw 0

include packet.asm
include print.asm
include byteswap.asm
include arp.asm
include ip.asm
include icmp.asm
include udp.asm
include intrhand.asm
include tcp.asm
include strings.inc
include buffers.inc

END entry_point
