handle dw 0

access_type:
  pushall
  cli
  mov ax, 0201h ;packet driver function 2
  mov bx, 0FFFFh ;all packet types
  mov cx, cs
  mov es, cx ;our receiver callback segment is our CS
  mov di, offset receiver_callback
  xor dl, dl ;null for interface number
  xor si, si ;null for type segment
  mov ds, si ;null for type pointer
  xor cx, cx ;null for type length
  int PACKETINT

  jc prepare_error_msg ;packet driver sets the carry flag on error returns
  mov cs:[handle], ax ;no error? then save our handle value from AX

  popall
  ret

prepare_error_msg:
	xor ch, ch
	mov cl, dh
	jmp error_exit

release_type:
  pushf
  cli
  mov ah, 3 ;packet driver function 3
  mov bx, cs:[handle]
  int PACKETINT
  popf
  ret

get_address:
  pushf
  cli
  push es
  push di
  push cx
  mov ax, cs
  mov es, ax
  mov ah, 6 ;packet driver function 6
  mov bx, cs:[handle]
  mov di, offset local_mac
  mov cx, 6
  int PACKETINT
  jc prepare_error_msg
  pop cx
  pop di
  pop es
  popf
  ret

drop_count dw 0
recv_count dw 0
buf_index  dw 0
receiver_callback:
  pushf
  cli ;re-entrancy is bad. very bad. the packet driver should have it
	  ;taken care of, but it's worth a clock tick to be sure
  cmp ax, 0
  jnz receiver_second ;if AX <> 0 then it's the second callback, data's ready

  cmp cs:[buf_index], INCOMING_BUFS ;do we have any available buffer memory?
  jnb receiver_drop ;if the buffers are all used, we have to drop the packet

  cmp cx, MTU+14 ;is it larger than our buffers can take?
  ja receiver_drop ;then we have to drop the packet

  mov cs:[recv_len], cx
  mov di, cs:[buf_index] ;first save packet size from CX into length array
  shl di, 1 ;each array index is a 2-byte word, double the buf_index value
  add di, offset recv_len ;and then sum it with the offset of the array
  mov cs:[di], cx ;and finally save the value

  push cs ;we need to return to the packet driver
  pop es  ;with our buffer pointer in ES:DI
  mov di, cs:[buf_index] ;get the current buffer index
  push ax
  push bx
  mov ax, (MTU+14) ;put the size of one buffer into AX
  mul di ;and multiply it with the current index value
  mov di, ax ;move it to DI
  add di, offset recv_bufs ;add the offset of the front of the buffer array
  pop bx
  pop ax

  inc cs:[buf_index] ;at this point we can increment buf_index
  inc cs:[recv_count]

  popf ;restore flags (do we even need to do this?)
  retf ;and go back to the packet driver and let it copy the data for us

receiver_second: ;the packet driver copied the packet into our buffer
  push di
  mov di, offset recv_type ;flip the byte order of the ethertype
  call swap_order_16       ;to little endian
  pop di

  ;call debug_recv
  call process_packet
  ;call crlf
  call rotate_packets

  popf
  retf

receiver_drop:
  xor di, di ;return null ES:DI
  mov es, di ;so the packet driver knows to drop our packet
  inc cs:[drop_count] ;and increment our dropped packet counter

  ;DEBUG: print notification of dropped packet
  ;mov si, offset strdroppkt
  ;call printmsg

  popf
  retf

rotate_packets:
  cld
  mov si, offset recv_bufs
  mov di, si
  add si, MTU+14

  mov ax, cs
  mov es, ax
  mov ds, ax
  xor ax, ax
  mov cx, INCOMING_BUFS-1
  add_packet_size:
	add ax, MTU+14
	loop add_packet_size
  shr ax, 1 ;we're going to MOVSW for speed, so divide byte count by two
  mov cx, ax ;rep count can only be in CX
  rep movsw ;go-go-gadget data copy

  dec cs:[buf_index] ;decrement current buffer index since we've made room
  ret

process_packet:
  cmp cs:[recv_type], 0806h
  jne not_arp
  call recv_arp
  ret

  not_arp:
  cmp cs:[recv_type], 0800h
  jne not_ip
  call recv_ip
  ret

  not_ip:
  ret

debug_recv:
  ;DEBUG: print notification of received packet
  mov si, offset strrecvpkt
  call printmsg
  push ax
  mov ax, cs:[recv_len]
  call print_dec
  pop ax
  mov si, offset strrecvpkt2
  call printmsg
  push dx
  mov dx, cs:[recv_type]
  call print_hex
  mov si, offset strrecvpkt3
  call printmsg

  mov si, offset strsource
  call printmsg
  mov si, offset recv_src
  call printmac
  mov si, offset strdestination
  call printmsg
  mov si, offset recv_dest
  call printmac
  call crlf
  pop dx
  ret

send_pkt:
  pushall
  cli
  mov ax, cs
  mov ds, ax
  add cx, 14
  mov si, offset pkt_dest
  mov ah, 4 ;packet driver function 4
  int PACKETINT
  popall
  ret

strrecvpkt  db '[DEBUG] Received ethernet packet! (', 0
strrecvpkt2 db ' bytes, ethertype: 0x',0
strrecvpkt3 db ')', 13, 10, 0
strdroppkt  db '[DEBUG] Dropped ethernet packet!', 13, 10, 0
strtype     db '    Frame type: ', 0

strtype_arp     db 'ARP', 13, 10, 0
strtype_ipv4    db 'IPv4 datagram', 13, 10, 0
strtype_unknown db 'Unknown', 13, 10, 0
strsource       db '    Source: ', 0
strdestination  db '        Destination: ', 0
