TCP_STATE_LISTEN      equ 1
TCP_STATE_SYN_SENT    equ 2
TCP_STATE_SYN_RECV    equ 3
TCP_STATE_ESTABLISHED equ 4
TCP_STATE_FIN_WAIT_1  equ 5
TCP_STATE_FIN_WAIT_2  equ 6
TCP_STATE_CLOSE_WAIT  equ 7
TCP_STATE_LAST_ACK    equ 8
TCP_STATE_TIME_WAIT   equ 9
TCP_STATE_CLOSED      equ 10
TCP_STATE_WAIT_ARP    equ 11

TCP_FLAG_ACK equ 00010000b
TCP_FLAG_PSH equ 00001000b
TCP_FLAG_RST equ 00000100b
TCP_FLAG_SYN equ 00000010b
TCP_FLAG_FIN equ 00000001b

TCP_RWIN    equ 1400
TCP_MAX_CON equ 10
TCP_TIMEOUT equ 3000 ;milliseconds to wait on ACK before we retransmit
TCP_RETRIES equ 15 ;how many times to resend/wait on ACK before aborting
TCP_WAIT    equ 10000 ;milliseconds until timeout with FIN and SYN packets
TCP_ARP_TIMEOUT equ 250 ;milliseconds between ARP broadcasts
TCP_ARP_RETRIES equ 10 ;how many times to broadcast ARP before giving up

tcp_srcport     dw 1
tcp_state       db TCP_MAX_CON dup (TCP_STATE_CLOSED)
tcp_retry       db TCP_MAX_CON dup (0)
tcp_peer_ip     db (TCP_MAX_CON*4) dup (0)
tcp_peer_mac    db (TCP_MAX_CON*6) dup (0FFh)
tcp_id_srcport  dw TCP_MAX_CON dup (0)
tcp_id_destport dw TCP_MAX_CON dup (0)
tcp_seq         dw (TCP_MAX_CON*2) dup (0)
tcp_ack         dw (TCP_MAX_CON*2) dup (0)
tcp_arp_state   db TCP_MAX_CON dup (0)
tcp_ack_wait    db TCP_MAX_CON dup (0)
tcp_send_seg    dw TCP_MAX_CON dup (0)
tcp_send_ptr    dw TCP_MAX_CON dup (0)
tcp_send_len    dw TCP_MAX_CON dup (0)
tcp_last_recv   dw (TCP_MAX_CON*2) dup (0)
tcp_is_new      db TCP_MAX_CON dup (0)
tcp_persistent  db TCP_MAX_CON dup (0) ;persistent listener flag
tcp_recvpause   db TCP_MAX_CON dup (1)

tcp_recv_size     dw 0
tcp_recv_data_len dw (TCP_MAX_CON) dup (0)

tcp_recv_real_off dw 0
tcp_recv_real_seq dw 2 dup (0)

tcp_recv_hdr:
tcp_recv_srcport  dw (TCP_MAX_CON) dup (0)
tcp_recv_destport dw (TCP_MAX_CON) dup (0)
tcp_recv_seq      dw 2 dup (0) ;32-bit value
tcp_recv_ack      dw 2 dup (0) ;32-bit value
tcp_recv_offset   db 0 ;shift right by 4 bits, then multiply by 4
tcp_recv_flags    db 0
tcp_recv_window   dw 0
tcp_recv_chksum   dw 0
tcp_recv_urgent   dw 0

tcp_pseudo_src    dw 2 dup (0)
tcp_pseudo_dest   dw 2 dup (0)
tcp_pseudo_zero   db 0
tcp_pseudo_prot   db 6
tcp_pseudo_length dw 0
tcp_send_hdr:
tcp_send_srcport  dw (TCP_MAX_CON) dup (0)
tcp_send_destport dw (TCP_MAX_CON) dup (0)
tcp_send_seq      dw 2 dup (0) ;32-bit value
tcp_send_ack      dw 2 dup (0) ;32-bit value
tcp_send_offset   db 0 ;shift right by 4 bits, then multiply by 4
tcp_send_flags    db 0
tcp_send_window   dw 2 ;512 in network-byte order
tcp_send_chksum   dw 0
tcp_send_urgent   dw 0
tcp_send_options  db 02h, 04h, 05h, 0B4h
tcp_send_data     db MTU dup (0)

tcp_cur_data_len  dw 0
tcp_cur_handle    dw 65535
tcp_temp32        dw 2 dup (0) ;used for processing of SYN and ACK numbers
tcp_recv:
  ;cli
  mov si, offset ip_data
  mov di, offset tcp_recv_hdr
  mov cx, 10 ;chunk of header we care about is the first 20 bytes
  rep movsw

  mov ax, cs:[ip_data_len]
  mov cs:[tcp_cur_data_len], ax
  mov al, byte ptr cs:[tcp_recv_hdr+12]
  and al, 11110000b
  xor ah, ah
  shr ax, 1 ;convert data offset field from DWORDs to bytes
  shr ax, 1
  mov cs:[tcp_recv_real_off], ax ;and store the result
  sub cs:[tcp_cur_data_len], ax

;  mov si, cs:[tcp_recv_real_off]
;  add si, offset ip_data
;  mov cx, cs:[tcp_cur_data_len]
;  cmp cx, 0
;  jz nodump
;  dumper:
;	lodsb
;	mov ah, 0Eh
;	int 10h
;	loop dumper
;  nodump:

  ;DEBUG printout
  ;mov si, offset strtcpdatalen
  ;call printmsg
  ;mov ax, cs:[tcp_cur_data_len]
  ;call print_dec
  ;call crlf

  ;figure out what to do with this packet...
  ;mov al, TCP_FLAG_PSH
  ;not al
  ;and byte ptr cs:[tcp_recv_hdr+13], al
  test byte ptr cs:[tcp_recv_hdr+13], TCP_FLAG_RST
  jnz tcp_got_rst2
  test byte ptr cs:[tcp_recv_hdr+13], TCP_FLAG_FIN
  jnz tcp_got_fin2
  test byte ptr cs:[tcp_recv_hdr+13], TCP_FLAG_SYN
  jnz tcp_got_syn
  test byte ptr cs:[tcp_recv_hdr+13], TCP_FLAG_ACK
  jnz tcp_got_ack2
  ret
tcp_got_rst2: ;gotta do this because the damn jumps are out of relative range
  call tcp_got_rst
  ret
tcp_got_ack2:
  call tcp_got_ack
  ret
tcp_got_fin2:
  call tcp_got_fin
  ret

save_old_handle5 dw 0
tcp_got_syn:
  test byte ptr cs:[tcp_recv_hdr+13], TCP_FLAG_ACK
  jz not_syn_ack
  jmp tcp_got_syn_ack
  not_syn_ack:
  pushall
  cld
  mov cx, TCP_MAX_CON
  mov ax, cs
  mov ds, ax
  mov es, ax
  ;mov si, offset save_ip
  ;call printip
  ;call crlf

  mov dx, word ptr cs:[tcp_recv_hdr+2]
  ;xchg dh, dl
  mov si, offset tcp_state
  ;mov cs:[save_bx], bx
  xor bx, bx
  find_new_listener:
	lodsb
	cmp al, TCP_STATE_LISTEN
	jne skip_port_check
	mov di, bx
	shl di, 1
	add di, offset tcp_id_srcport
	cmp dx, word ptr cs:[di]
	je found_listener
	skip_port_check:
	inc bx
	loop find_new_listener
  popall
  mov bx, 65535
  ret
  found_listener:
	mov dx, cs:[tcp_cur_handle]
	mov cs:[save_old_handle5], dx
	mov cs:[tcp_cur_handle], bx
	mov di, bx
	shl di, 1
	add di, offset tcp_id_destport
	mov ax, word ptr cs:[tcp_recv_hdr]
	mov cs:[di], ax

	mov di, bx
	shl di, 1
	shl di, 1
	add di, offset tcp_peer_ip
	mov ax, word ptr cs:[ip_src]
	mov cs:[di], ax
	mov ax, word ptr cs:[ip_src+2]
	mov cs:[di+2], ax

	mov si, offset recv_src ;sender's MAC
	mov di, offset tcp_peer_mac-6
	mov cx, cs:[tcp_cur_handle]
	inc_mac_index3:
	  add di, 6
	  loop inc_mac_index3
	mov cx, 6
	rep movsb

	mov di, cs:[tcp_cur_handle]
	shl di, 1
	shl di, 1
	add di, offset tcp_ack
	mov ax, word ptr cs:[tcp_recv_hdr+4]
	mov cs:[di], ax
	mov ax, word ptr cs:[tcp_recv_hdr+6]
	mov cs:[di+2], ax
	inc_seq
	inc_ack

	;call tcp_reset_timer_and_retry
	mov cl, TCP_STATE_SYN_RECV
	mov di, cs:[tcp_cur_handle]
	add di, offset tcp_state
	mov byte ptr cs:[di], cl
	call tcp_change_state
	call tcp_send_syn_ack
	inc_seq

	mov dx, cs:[save_old_handle5]
	mov cs:[tcp_cur_handle], dx

  ;DEBUG printout
  ;mov si, offset strlisten
  ;call printmsg
  popall
  just_ret2:
	ret

tcp_got_syn_ack:
  call match_ident ;will put handle number in tcp_cur_handle
  jc just_ret2 ;or return with carry set if no handle match found

  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_state
  cmp byte ptr cs:[di], TCP_STATE_SYN_SENT
  jne just_ret2 ;maybe send a RST in this case?
  mov byte ptr cs:[di], TCP_STATE_ESTABLISHED

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  add di, offset tcp_ack
  mov ax, word ptr cs:[tcp_recv_hdr+4]
  mov cs:[di], ax
  mov ax, word ptr cs:[tcp_recv_hdr+6]
  mov cs:[di+2], ax
  inc_ack ;macro
  ;inc_seq ;macro

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  add di, offset tcp_seq
  mov ax, word ptr cs:[tcp_recv_hdr+8]
  mov cs:[di], ax
  mov ax, word ptr cs:[tcp_recv_hdr+10]
  mov cs:[di+2], ax

  call tcp_send_empty_ack

  mov cl, TCP_STATE_ESTABLISHED
  call tcp_change_state
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_ack_wait
  mov byte ptr cs:[di], 0 ;clear ACK wait flag, it gets set when sending data packets
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_is_new
  mov byte ptr cs:[di], 1

  ;DEBUG printout
  ;mov si, offset strsynack
  ;call printmsg
  just_ret:
	ret

temp_ack_1 dw 0
temp_ack_2 dw 0
tcp_cb_len dw 0
tcp_got_ack:
  call match_ident
  jc just_ret

  ;mov di, cs:[tcp_cur_handle]
  ;add di, offset tcp_recvpause
  ;cmp byte ptr cs:[di], 0
  ;jz just_ret

  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_state
  cmp byte ptr cs:[di], TCP_STATE_SYN_RECV
  jne not_accept_request
  jmp tcp_accept

  not_accept_request:
  mov ax, word ptr cs:[tcp_recv_hdr+4]
  xchg ah, al
  mov cs:[temp_ack_2], ax
  mov ax, word ptr cs:[tcp_recv_hdr+6]
  xchg ah, al
  mov cs:[temp_ack_1], ax

  inc_ack

  mov ax, cs:[tcp_cur_data_len]
  mov cs:[tcp_cb_len], ax

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  add di, offset tcp_ack
  mov ax, word ptr cs:[tcp_recv_hdr+4]
  xchg ah, al
  cmp cs:[temp_ack_2], ax
  jna no_abort ;tcp_abort ;if their ACK number is higher than our seq, bad news...
  jmp tcp_abort
  no_abort:
  jnb no_retrans ;tcp_retransmit2 ;if lower, they missed something... retransmit
  jmp tcp_retransmit
  no_retrans:
  xchg ah, al
  mov cs:[di], ax
  mov ax, word ptr cs:[tcp_recv_hdr+6]
  mov cs:[di+2], ax
  xchg ah, al
  cmp cs:[temp_ack_1], ax
  jna no_abort2 ;tcp_abort ;if their ACK number is higher than our seq, bad news...
  jmp tcp_abort
  no_abort2:
  jnb no_retrans2 ;tcp_retransmit ;if lower, they missed something... retransmit
  jmp tcp_retransmit
  no_retrans2:
  xchg ah, al

  ;test byte ptr cs:[di], TCP_FLAG_PSH
  mov cx, cs:[tcp_cb_len]
  cmp cx, 0
  jz skip_len_loop
  mov ax, cs:[di+2]
  mov bx, cs:[di]
  xchg ah, al
  xchg bh, bl
  add ax, cx
  adc bx, 0
  xchg ah, al
  xchg bh, bl
  mov cs:[di+2], ax
  mov cs:[di], bx


  ;add_len_loop: ;this could be done WAY faster... fix it, you lazy bum!
  ;	add byte ptr cs:[di+3], 1
  ;	adc byte ptr cs:[di+2], 0
  ;	adc byte ptr cs:[di+1], 0
  ;	adc byte ptr cs:[di], 0
  ;	loop add_len_loop
  skip_len_loop:

  ;mov di, cs:[tcp_cur_handle]
  ;shl di, 1
  ;add di, offset tcp_send_len
  ;mov word ptr cs:[di], 0 ;not waiting for an ACK anymore

  ;test byte ptr cs:[tcp_recv_hdr+13], TCP_FLAG_PSH
  ;jne do_not_ack
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_ack_wait
  mov byte ptr cs:[di], 0 ;not waiting for an ACK anymore

  ;mov cx, cs:[tcp_cur_data_len]
  ;mov cx, cs:[tcp_cb_len]
  cmp cs:[tcp_cb_len], 0 ;don't send another ACK back if no new data
  jz no_ack_back ;do_not_ack ;just_ret3

  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_is_new
  cmp byte ptr cs:[di], 1
  je skip_recv_ack_check

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  add di, offset tcp_last_recv
  mov ax, word ptr cs:[tcp_recv_hdr+4]
  cmp ax, cs:[di]
  jne tcp_recv_skip
  mov ax, word ptr cs:[tcp_recv_hdr+6]
  cmp ax, cs:[di+2]
  jne tcp_recv_skip

  skip_recv_ack_check:
  do_not_ack:
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_is_new
  mov byte ptr cs:[di], 0

  mov ax, cs:[tcp_callback_ptr]
  or ax, cs:[tcp_callback_seg]
  jz tcp_recv_skip

  mov si, offset ip_data
  add si, cs:[tcp_recv_real_off]
  mov cx, cs:[tcp_cb_len]
  mov ah, 1 ;callback function 1 means data receive
  call tcp_callback
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_recvpause
  mov byte ptr cs:[di], al
  cmp al, 1
  jne no_ack_back

  call tcp_send_empty_ack

  no_ack_back:
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  mov si, di
  add di, offset tcp_ack
  add si, offset tcp_last_recv
  mov ax, cs:[di]
  mov cs:[si], ax
  mov ax, cs:[di+2]
  mov cs:[si+2], ax

  tcp_recv_skip:
  ;mov si, offset ip_data
  ;add si, cs:[tcp_recv_real_off]
  ;mov cx, cs:[tcp_cur_data_len]
  ;dump_data:
  ;	lodsb
  ;	mov ah, 0Eh
  ;	int 10h
  ;	skip_byte_dump:
  ;	loop dump_data

  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_state
  cmp byte ptr cs:[di], TCP_STATE_FIN_WAIT_1
  je teardown
  cmp byte ptr cs:[di], TCP_STATE_FIN_WAIT_2
  je teardown
  jmp not_teardown

  teardown:
  call tcp_reset
  ret

  not_teardown:
  ;DEBUG printout
  ;mov si, offset strack
  ;call printmsg
  test byte ptr cs:[tcp_recv_hdr+13], TCP_FLAG_FIN
  jz no_fin
  call tcp_send_fin_ack
  no_fin:
  ret

tcp_accept:
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_is_new
  mov byte ptr cs:[di], 1
  mov cl, TCP_STATE_ESTABLISHED
  call tcp_change_state
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_state
  mov byte ptr cs:[di], cl
  ret

tcp_abort:
  pushf
  pushall
  call tcp_send_rst
  call tcp_reset
  popall
  popf
  ret

tcp_retransmit:
  ;pushall
  ;call tcp_data_out
  ;mov si, offset strretrans
  ;call printmsg
  ;popall
  ret

tcp_got_fin:
  call match_ident ;will return handle in BX, or set carry if no match
  jc just_ret3
  inc_ack ;macro

  ;respond with ACK, FIN+ACK to finalize the connection teardown
  call tcp_send_empty_ack
  call tcp_send_fin_ack
  call tcp_reset
  ret

tcp_got_rst:
  call match_ident ;will return handle in BX, or set carry if no match
  jc just_ret3

  call tcp_reset

  ;DEBUG printout
  ;mov si, offset strrst
  ;call printmsg
  just_ret3:
	ret

save_old_handle2 dw 0
tcp_init_normal_close:
  pushall
  mov bx, cs:[tcp_cur_handle]
  mov cs:[save_old_handle2], bx

  mov cl, TCP_STATE_FIN_WAIT_1
  call tcp_change_state
  call tcp_send_fin_ack

  mov bx, cs:[save_old_handle2]
  mov cs:[tcp_cur_handle], bx
  popall
  ret

tcp_send_rst:
  mov di, offset tcp_send_srcport
  mov bx, cs:[tcp_id_srcport]
  xchg bh, bl
  mov cs:[di], bx ;set source port in send header
  mov bx, cs:[new_port]
  xchg bh, bl
  mov cs:[di+2], bx ;set dest port in send header
  mov byte ptr cs:[tcp_send_hdr+13], TCP_FLAG_RST
  mov cs:[tcp_no_data], 1
  mov di, offset tcp_send_data ;clear first 2 bytes of data area (checksum)
  xor ax, ax
  stosw
  ;mov cs:[tcp_cur_data_len], 0 ;clear this or problems will occur!
  xor cx, cx ;zero bytes of data
  mov si, cs:[tcp_cur_handle]
  shl si, 1
  shl si, 1
  add si, offset tcp_peer_ip
  call tcp_send
  ret

tcp_no_data db 0
tcp_send_empty_ack:
  mov di, offset tcp_send_srcport
  mov bx, cs:[tcp_id_srcport]
  xchg bh, bl
  mov cs:[di], bx ;set source port in send header
  mov bx, cs:[new_port]
  xchg bh, bl
  mov cs:[di+2], bx ;set dest port in send header
  mov byte ptr cs:[tcp_send_hdr+13], TCP_FLAG_ACK
  mov cs:[tcp_no_data], 1
  mov di, offset tcp_send_data ;clear first 2 bytes of data area (checksum)
  xor ax, ax
  stosw

  ;mov cs:[tcp_cur_data_len], 0 ;clear this or problems will occur!
  xor cx, cx ;zero bytes of data
  mov si, cs:[tcp_cur_handle]
  shl si, 1
  shl si, 1
  add si, offset tcp_peer_ip
  call tcp_send
  ret

tcp_send_fin_ack:
  mov di, offset tcp_send_srcport
  mov bx, cs:[tcp_id_srcport]
  xchg bh, bl
  mov cs:[di], bx ;set source port in send header
  mov bx, cs:[new_port]
  xchg bh, bl
  mov cs:[di+2], bx ;set dest port in send header
  mov byte ptr cs:[tcp_send_hdr+13], TCP_FLAG_ACK
  or byte ptr cs:[tcp_send_hdr+13], TCP_FLAG_FIN
  mov cs:[tcp_no_data], 1
  ;mov cs:[tcp_cur_data_len], 0 ;clear this or problems will occur!
  xor cx, cx ;zero bytes of data
  mov si, cs:[tcp_cur_handle]
  shl si, 1
  shl si, 1
  add si, offset tcp_peer_ip
  call tcp_send
  mov cl, TCP_STATE_FIN_WAIT_2
  call tcp_change_state
  ret

cur_match_test dw 0
match_flag     db 1
match_ident:
  pushall
  cld
  mov ax, cs
  mov ds, ax
  mov es, ax

  mov cx, TCP_MAX_CON
  mov cs:[cur_match_test], 0
  match_ident_loop:
	mov cs:[match_flag], 1
	mov di, cs:[cur_match_test]
	shl di, 1
	shl di, 1
	add di, offset tcp_peer_ip
	mov si, offset ip_src
	call does_ip_match
	jc no_match

	mov di, cs:[cur_match_test]
	shl di, 1
	add di, offset tcp_id_srcport
	mov si, offset tcp_recv_hdr+2 ;destport
	cmpsw
	jne no_match

	mov di, cs:[cur_match_test]
	shl di, 1
	add di, offset tcp_id_destport
	mov si, offset tcp_recv_hdr ;tcp_recv_srcport
	cmpsw
	jne no_match
	jmp match_ident_okay

  no_match:
	inc cs:[cur_match_test]
  loop match_ident_loop

  match_ident_fail:
	popall
	stc
	ret
  match_ident_okay:
	mov bx, cs:[cur_match_test]
	mov cs:[tcp_cur_handle], bx
	;DEBUG printout
	;mov si, offset strmatchid
	;call printmsg
	mov ax, cs:[tcp_cur_handle]
	;call print_dec
	;call crlf
	popall
	clc
	ret

tcp_reset: ;reset a TCP connection handle
  pushall
  mov cl, TCP_STATE_CLOSED
  call tcp_change_state
  mov dx, cs:[tcp_cur_handle]

  mov bx, offset tcp_state
  mov di, dx
  mov byte ptr cs:[bx+di], TCP_STATE_CLOSED

  mov bx, offset tcp_retry
  mov byte ptr cs:[bx+di], 0

  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_recvpause
  mov byte ptr cs:[di], 1

  mov di, cs:[tcp_recv_data_len]
  shl di, 1
  add di, offset tcp_recv_data_len
  mov word ptr cs:[di], 0

  mov di, dx
  shl di, 1
  shl di, 1
  add di, offset tcp_peer_ip
  mov word ptr cs:[di], 0

  mov di, dx
  shl di, 1
  add di, offset tcp_id_srcport
  mov word ptr cs:[di], 0

  mov di, dx
  shl di, 1
  add di, offset tcp_id_destport
  mov word ptr cs:[di], 0

  mov bx, offset tcp_seq
  mov di, dx
  shl di, 1
  shl di, 1
  mov word ptr cs:[bx+di], 0
  add di, 2
  mov word ptr cs:[bx+di], 0

  mov bx, offset tcp_ack
  sub di, 2
  mov word ptr cs:[bx+di], 0
  add di, 2
  mov word ptr cs:[bx+di], 0

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  add di, offset tcp_last_recv
  mov word ptr cs:[di], 65535
  mov word ptr cs:[di+2], 65535

  ;DEBUG printout
  ;mov si, offset strclosed
  ;call printmsg
  mov ax, cs:[tcp_cur_handle]
  ;call print_dec
  ;call crlf
  popall
  ret

save_bx dw 0
new_port dw 0
save_ip dw 2 dup (0)
save_hand dw 0
make_listener db 0
temp_persistent db 0
tcp_connect:
  cld
  pushall
  mov cs:[temp_persistent], al
  lodsw
  mov cs:[save_ip], ax
  lodsw
  mov cs:[save_ip+2], ax
  mov cs:[new_port], bx

  mov cx, TCP_MAX_CON
  mov ax, cs
  mov ds, ax
  mov es, ax
  ;mov si, offset save_ip
  ;call printip
  ;call crlf

  mov si, offset tcp_state
  mov cs:[save_bx], bx
  xor bx, bx
  find_new_handle:
	lodsb
	cmp al, TCP_STATE_CLOSED
	je found_new_handle
	inc bx
	loop find_new_handle
  popall
  mov bx, 65535
  ret
  found_new_handle:
	mov cs:[save_hand], bx
	mov cs:[tcp_cur_handle], bx
	;call tcp_reset_timer_and_retry
	;call tcp_reset
	cmp cs:[make_listener], 1
	jne tcp_not_listen
	jmp tcp_open_listen

	tcp_not_listen:
	mov di, offset tcp_persistent
	mov byte ptr cs:[bx+di], 0
	mov si, offset save_ip
	mov di, bx
	shl di, 1
	shl di, 1
	add di, offset tcp_peer_ip
	movsw
	movsw

	mov ax, cs:[tcp_srcport]
	xchg ah, al
	mov di, cs:[save_hand]
	shl di, 1
	add di, offset tcp_id_srcport
	mov cs:[di], ax

	mov ax, cs:[new_port]
	xchg ah, al
	mov di, cs:[save_hand]
	shl di, 1
	add di, offset tcp_id_destport
	mov cs:[di], ax ;stosw ;movsw

	mov di, cs:[tcp_cur_handle]
	shl di, 1
	shl di, 1
	add di, offset tcp_seq
	mov word ptr cs:[di], 0
	mov word ptr cs:[di+2], 0

	mov di, cs:[tcp_cur_handle]
	shl di, 1
	shl di, 1
	add di, offset tcp_ack
	mov word ptr cs:[di], 0
	mov word ptr cs:[di+2], 0

	call tcp_subnet_check

	;call crlf
	mov bx, cs:[save_hand]
	call tcp_send_syn
	mov ax, cs:[save_hand] ;bx

	;call print_dec
	;call crlf
	mov si, offset tcp_peer_ip
	;call printip
	;call crlf
	mov si, cs:[save_hand]
	shl si, 1
	add si, offset tcp_id_srcport
	lodsw
	xchg ah, al
	;call print_dec
	;call crlf
	mov si, cs:[save_hand]
	shl si, 1
	add si, offset tcp_id_destport
	lodsw
	xchg ah, al
	;call print_dec
	;call crlf

	popall
	mov bx, cs:[tcp_cur_handle]
	ret

tcp_open_relisten:
	cld
	pushall
	mov cs:[temp_persistent], 1
tcp_open_listen:
	mov cs:[make_listener], 0
	mov di, cs:[tcp_cur_handle]
	add di, offset tcp_persistent
	mov al, cs:[temp_persistent]
	mov cs:[di], al

	mov di, cs:[save_hand]
	add di, offset tcp_state
	mov byte ptr cs:[di], TCP_STATE_LISTEN

	mov ax, cs:[new_port]
	mov di, cs:[save_hand]
	shl di, 1
	add di, offset tcp_id_srcport
	xchg ah, al
	mov cs:[di], ax

	mov cl, TCP_STATE_LISTEN
	call tcp_change_state

	popall
	mov bx, cs:[tcp_cur_handle]
	ret

tcp_change_state: ;enter with new state in CL
  ;cli
  push di
  mov di, offset tcp_state
  add di, cs:[tcp_cur_handle]
  mov byte ptr cs:[di], cl
  ;cmp cl, TCP_STATE_CLOSED
  ;je skip_reset_timer
  ;cmp cl, TCP_STATE_ESTABLISHED
  ;je skip_reset_timer
  ;cmp cl, TCP_STATE_LISTEN
  ;je skip_reset_timer

  ;mov di, cs:[tcp_cur_handle]
  ;shl di, 1
  ;add di, offset timers
  ;mov word ptr cs:[di], 0 ;reset timeout counter

  skip_reset_timer:
  pop di
  call tcp_reset_timer_and_retry
  call tcp_status_callback
  ;DEBUG printout
  ;mov si, offset strchange
  ;call printmsg
  mov al, cl
  xor ah, ah
  ;call print_dec
  ;call crlf
  ret

tcp_send_syn:
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_state
  cmp byte ptr cs:[di], TCP_STATE_WAIT_ARP
  je just_ret4 ;not ready to continue yet...

  xor cx, cx ;zero bytes of data
  mov byte ptr cs:[tcp_send_hdr+13], TCP_FLAG_SYN
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  add di, offset tcp_seq
  mov word ptr cs:[di], 12345
  mov word ptr cs:[di+2], 6789

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  add di, offset tcp_ack
  mov word ptr cs:[di], 0
  mov word ptr cs:[di+2], 0
  mov cs:[tcp_no_data], 1
  mov di, offset tcp_send_data ;clear first 2 bytes of data area (checksum)
  xor ax, ax
  stosw

  mov si, offset save_ip
  call tcp_send
  mov cl, TCP_STATE_SYN_SENT
  call tcp_change_state
  just_ret4:
  ret

tcp_send_syn_ack:
  ;mov cl, TCP_STATE_ESTABLISHED
  ;call tcp_change_state
  mov di, offset tcp_send_hdr ;tcp_send_srcport
  mov bx, cs:[tcp_id_srcport]
  ;xchg bh, bl
  mov cs:[di], bx ;set source port in send header
  ;inc cs:[tcp_srcport]
  mov si, cs:[tcp_cur_handle]
  shl si, 1
  add si, offset tcp_id_destport
  mov bx, cs:[si]
  xchg bh, bl
  mov cs:[di+2], bx ;set dest port in send header
  mov byte ptr cs:[tcp_send_hdr+13], TCP_FLAG_SYN
  or byte ptr cs:[tcp_send_hdr+13], TCP_FLAG_ACK
  mov cs:[tcp_no_data], 1
  mov di, offset tcp_send_data ;clear first 2 bytes of data area (checksum)
  xor ax, ax
  stosw
  xor cx, cx ;zero bytes of data
  ;mov si, offset save_ip
  mov si, cs:[tcp_cur_handle]
  shl si, 1
  shl si, 1
  add si, offset tcp_peer_ip

  cld
  mov ax, cs
  mov ds, ax
  mov es, ax
  call tcp_send
  ret

tcp_data_out:
  cli
  cld
  mov ax, cs
  mov es, ax
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset tcp_send_len
  mov cx, cs:[di]

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset tcp_send_seg
  mov bx, cs:[di]
  mov ds, bx
  mov si, cs:[tcp_cur_handle]
  shl si, 1
  add si, offset tcp_send_ptr
  mov si, cs:[si]
  mov di, offset tcp_send_hdr+24
  ;push cx
  ;push si
  ;dump_send:
  ;	lodsb
  ;	mov ah, 0Eh
  ;	int 10h
  ;	loop dump_send
  ;pop si
  ;pop cx
  rep movsb
  mov al, 0
  stosb
  stosb
  stosb
  stosb
  mov ax, cs
  mov ds, ax

  mov di, offset tcp_send_srcport
  mov bx, cs:[tcp_id_srcport]
  xchg bh, bl
  mov cs:[di], bx ;set source port in send header
  mov bx, cs:[new_port]
  xchg bh, bl
  mov cs:[di+2], bx ;set dest port in send header
  mov byte ptr cs:[tcp_send_hdr+13], TCP_FLAG_ACK
  or byte ptr cs:[tcp_send_hdr+13], TCP_FLAG_PSH

  mov cs:[tcp_cur_data_len], 0 ;otherwise their ACK back would break stuff

  mov si, cs:[tcp_cur_handle]
  shl si, 1
  shl si, 1
  add si, offset tcp_peer_ip
  call tcp_send
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_ack_wait
  mov byte ptr cs:[di], 1 ;set ACK wait to true HEY MIKE YOU CHANGED THIS }:A
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset tcp_send_len
  mov word ptr cs:[di], 0 ;set ACK wait to true
  ;call tcp_reset_timer_and_retry
  ret

tcp_send: ;enter this with destination IP pointer in DS:SI
  mov ax, cs
  mov es, ax
  mov di, offset ip_send_dest
  movsw
  movsw

  mov ax, cs
  mov ds, ax
  mov si, offset tcp_peer_mac-6
  mov cx, cs:[tcp_cur_handle]
  inc_mac_loop3:
	add si, 6
	loop inc_mac_loop3
  mov di, offset pkt_dest
  mov cx, 3
  rep movsw

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset tcp_id_srcport ;cs:[tcp_send_srcport]
  mov ax, cs:[di]
  mov word ptr cs:[tcp_send_hdr], ax

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset tcp_id_destport ;cs:[tcp_send_srcport]
  mov ax, cs:[di]
  mov word ptr cs:[tcp_send_hdr+2], ax

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  add di, offset tcp_seq
  mov ax, cs:[di]
  mov word ptr cs:[tcp_send_hdr+4], ax
  mov ax, cs:[di+2]
  mov word ptr cs:[tcp_send_hdr+6], ax

  mov di, cs:[tcp_cur_handle]
  shl di, 1
  shl di, 1
  add di, offset tcp_ack
  mov ax, cs:[di]
  mov word ptr cs:[tcp_send_hdr+8], ax
  mov ax, cs:[di+2]
  mov word ptr cs:[tcp_send_hdr+10], ax

  mov di, offset tcp_send_hdr+14
  mov ax, TCP_RWIN
  xchg ah, al
  mov word ptr cs:[di], ax

  xor cx, cx
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_ack_wait
  cmp byte ptr cs:[di], 1
  je skip_send_check
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset tcp_send_len
  mov cx, cs:[di]
  skip_send_check:
  add cx, 24 ;add size of TCP header to length count in CX
  mov cs:[ip_send_total_len], cx ;put the length into the IP send packet hdr
  mov byte ptr cs:[tcp_send_hdr+12], 01100000b ;data offset = 20/4 = 5 in high nibble
  mov word ptr cs:[tcp_send_hdr+20], 0402h ;option 2 = MSS, length = 4
  mov word ptr cs:[tcp_send_hdr+22], ax ;MSS=RWIN

  call tcp_calc_chksum

  mov ax, cs
  mov ds, ax
  mov es, ax
  mov di, offset ip_send_data
  mov si, offset tcp_send_hdr
  rep movsb

  mov cs:[ip_send_protocol], 6

  mov cs:[ip_send_is_tcp], 1
  call ip_send
  ;mov cs:[tcp_cur_data_len], 0 ;clear this or problems will occur @ next ACK!
  mov di, offset tcp_ack_wait
  add di, cs:[tcp_cur_handle]
  mov byte ptr cs:[di], 0 ;remove this later when you fix shit...
  mov cs:[tcp_no_data], 0
  ret

tcp_real_send_len dw 0
tcp_calc_chksum:
  ;push cx
  mov ax, cs
  mov ds, ax
  mov es, ax
  mov ax, word ptr [local_ip]
  mov [tcp_pseudo_src], ax
  mov ax, word ptr [local_ip+2]
  mov [tcp_pseudo_src+2], ax
  mov ax, word ptr [ip_send_dest]
  mov [tcp_pseudo_dest], ax
  mov ax, word ptr [ip_send_dest+2]
  mov [tcp_pseudo_dest+2], ax
  ;mov di, cs:[tcp_cur_handle]
  ;shl di, 1
  ;shl di, 1
  ;add di, offset tcp_peer_ip
  ;mov ax, cs:[di]
  ;mov cs:[tcp_pseudo_dest], ax
  ;mov ax, cs:[di+2]
  ;mov cs:[tcp_pseudo_dest+2], ax
  xchg ch, cl
  mov [tcp_pseudo_length], cx
  xchg ch, cl

  xor bx, bx ;BX will be our accumulator when calculating checksum
  mov word ptr cs:[tcp_send_hdr+16], 0 ;clear checksum field when calculating
  mov si, offset tcp_pseudo_src
  ;mov cs:[tcp_real_send_len], cx
  push cx
  shr cx, 1
  add cx, 6
  cmp cs:[tcp_no_data], 1
  je tcp_skip_cx_send_inc
  inc cx
  ;test cs:[tcp_real_send_len], 1
  ;jnz tcp_send_not_odd
  ;inc cx
  tcp_skip_cx_send_inc:
  mov cs:[tcp_no_data], 0
  clc
  loop_calc_tcp_chksum:
	lodsw
	xchg ah, al ;byteswap it
	adc bx, ax
	loop loop_calc_tcp_chksum
  adc bx, 0 ;account for any left over carry bit
  not bx ;one's complement the accumulated value
  xchg bh, bl
  mov word ptr [tcp_send_hdr+16], bx ;insert checksum into TCP header
  pop cx
  ret


tcp_status_callback:
  ;cli
  mov ax, 0 ;callback function 0 = status update
  mov bx, cs:[tcp_cur_handle] ;AH = callback function, BX = handle, CL = new status
  jmp tcp_callback

temp_ax dw 0
temp_bx dw 0
tcp_callback:
  mov cs:[temp_ax], ax
  mov cs:[temp_bx], bx
  mov ax, cs:[tcp_callback_ptr]
  or ax, cs:[tcp_callback_seg]
  jz tcp_callback_done

  mov bx, cs:[tcp_cur_handle]
  pushf
  mov ax, offset tcp_callback_done
  push cs
  push ax
  mov ax, cs:[temp_ax]

  db 0EAh
  tcp_callback_ptr dw 0
  tcp_callback_seg dw 0

  tcp_callback_done:
	mov bx, cs:[temp_bx]
	ret

tcp_subnet_temp  db 4 dup (0)
tcp_subnet_temp2 db 4 dup (0)
tcp_subnet_check:
  cld
  mov si, cs:[tcp_cur_handle]
  shl si, 1
  shl si, 1
  add si, offset tcp_peer_ip
  mov di, offset tcp_subnet_temp
  mov cx, 4
  rep movsb
  mov si, offset local_ip
  mov di, offset tcp_subnet_temp2
  mov cx, 4
  rep movsb
  mov si, offset subnet
  mov di, offset tcp_subnet_temp
  mov cx, 4
  tcp_subnet_loop:
	lodsb
	and al, [di]
	stosb
	loop tcp_subnet_loop
  mov si, offset subnet
  mov di, offset tcp_subnet_temp2
  mov cx, 4
  tcp_subnet_loop2:
	lodsb
	and al, [di]
	stosb
	loop tcp_subnet_loop2
  mov si, offset tcp_subnet_temp
  mov di, offset tcp_subnet_temp2
  mov cx, 4
  tcp_subnet_loop3:
	cmpsb
	jne tcp_subnet_is_remote
	loop tcp_subnet_loop3
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_state
  mov byte ptr cs:[di], TCP_STATE_WAIT_ARP
  call tcp_initiate_arp
  ret
tcp_subnet_is_remote:
  mov di, offset tcp_peer_mac-6
  mov cx, cs:[tcp_cur_handle]
  inc_mac_index2:
	add di, 6
	loop inc_mac_index2
  mov si, offset gateway_mac
  mov cx, 3
  rep movsw
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_state
  mov byte ptr cs:[di], TCP_STATE_CLOSED ;just to be sure before sending SYN
  ret

tcp_arp_dest      db 0FFh, 0FFh, 0FFh, 0FFh, 0FFh, 0FFh ;broadcast packet
tcp_arp_src       db 6 dup (0)
tcp_arp_type      dw 0608h ;ethertype 0x0806 in network-byte order
tcp_arp_hw        dw 0100h ;hardware protocol 1 in network-byte order (ethernet)
tcp_arp_prot      dw 0008h ;protocol 0x0800 in network-byte order (IP)
tcp_arp_adr_len   db 6, 4
tcp_arp_oper      dw 0100h ;ARP operation 1 = request
tcp_arp_sender_hw db 6 dup (0)
tcp_arp_sender_ip db 4 dup (0)
tcp_arp_target_hw db 6 dup (0)
tcp_arp_target_ip db 4 dup (0)

tcp_initiate_arp:
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_arp_state
  mov byte ptr cs:[di], TCP_ARP_RETRIES

tcp_send_arp:
  pushall
  cld
  mov ax, cs
  mov ds, ax
  mov es, ax
  ;lets form the ARP request packet
  mov si, offset local_mac
  mov di, offset tcp_arp_src
  mov cx, 3
  rep movsw
  mov si, offset local_mac
  mov di, offset tcp_arp_sender_hw
  mov cx, 3
  rep movsw
  mov si, offset local_ip
  mov di, offset tcp_arp_sender_ip
  mov cx, 2
  rep movsw
  mov si, cs:[tcp_cur_handle]
  shl si, 1
  shl si, 1
  add si, offset tcp_peer_ip
  mov di, offset tcp_arp_target_ip
  mov cx, 2
  rep movsw

  ;;ARP packet is ready
  ;mov di, cs:[tcp_cur_handle]
  ;shl di, 1
  ;add di, offset timers
  ;mov word ptr cs:[di], 0 ;reset the handle's timer
  ;call tcp_reset_timer_and_retry
  pushf
  cli
  mov ax, cs
  mov ds, ax
  add cx, 42 ;total of 42 bytes
  mov si, offset tcp_arp_dest
  mov ah, 4 ;packet driver function 4
  int PACKETINT
  popf
  popall
  ret

tcp_check_arp_reply:
  mov bx, cs:[tcp_cur_handle]
  mov cs:[save_old_handle], bx
  cld
  mov bx, 0 ;start checking IPs at index 0 of tcp_peer_ip array
  check_arp_reply_loop2:
	mov di, offset tcp_state
	cmp byte ptr cs:[bx+di], TCP_STATE_WAIT_ARP
	jne tcp_check_arp_next_loop ;skip this index if isn't waiting on ARP
	mov di, bx
	shl di, 1
	shl di, 1
	add di, offset tcp_peer_ip
	mov si, offset recv_data+14 ;sender's IP
	mov cx, 4
	check_arp_reply_inner_loop:
	  cmpsb
	  jne tcp_check_arp_next_loop
	  loop check_arp_reply_inner_loop
	jmp tcp_arp_found
tcp_check_arp_next_loop:
	inc bx
	cmp bx, TCP_MAX_CON
	je tcp_check_arp_done
	jmp check_arp_reply_loop2
tcp_arp_found:
  mov cs:[tcp_cur_handle], bx
  mov si, offset recv_data+8 ;sender's MAC
  mov di, offset tcp_peer_mac-6
  mov cx, cs:[tcp_cur_handle]
  inc_mac_index:
	add di, 6
	loop inc_mac_index
  mov cx, 6
  rep movsb
  mov di, offset tcp_state
  mov byte ptr cs:[bx+di], TCP_STATE_CLOSED
  call tcp_send_syn

tcp_check_arp_done:
  mov bx, cs:[save_old_handle]
  mov cs:[tcp_cur_handle], bx
  ret

save_old_handle dw 0
save_old_handle3 dw 0

tcp_process_tick:
  pushall
  mov bx, cs:[tcp_cur_handle]
  mov cs:[save_old_handle3], bx

  xor bx, bx
  mov di, offset tcp_state
  process_tick_loop:
	mov cs:[tcp_cur_handle], bx
	cmp byte ptr cs:[bx+di], TCP_STATE_CLOSED
	jne next_1
	jmp process_tick_loop_return
	next_1:
	cmp byte ptr cs:[bx+di], TCP_STATE_LISTEN
	jne next_2
	jmp process_tick_loop_return
	next_2:
	;cmp byte ptr cs:[bx+di], TCP_STATE_ESTABLISHED
	;je process_established
	cmp byte ptr cs:[bx+di], TCP_STATE_FIN_WAIT_1
	je process_fin_wait
	cmp byte ptr cs:[bx+di], TCP_STATE_FIN_WAIT_2
	je process_fin_wait
	;cmp byte ptr cs:[bx+di], TCP_STATE_SYN_SENT
	;je process_syn_wait2
	;cmp byte ptr cs:[bx+di], TCP_STATE_SYN_RECV
	;jne process_syn_wait2
	cmp byte ptr cs:[bx+di], TCP_STATE_WAIT_ARP
	jne process_tick_loop_return
	jmp process_arp_wait

  process_tick_loop_return:
	inc bx
	cmp bx, TCP_MAX_CON
	jne process_tick_loop

  mov bx, cs:[save_old_handle3]
  mov cs:[tcp_cur_handle], bx
  popall
  ret

process_syn_wait2:
  jmp process_syn_wait

process_arp_wait2:
  jmp process_arp_wait

process_established:
  mov di, offset tcp_ack_wait
  cmp byte ptr cs:[bx+di], 0
  jz process_tick_loop_return
  mov di, bx
  shl di, 1
  add di, offset timers
  cmp word ptr cs:[di], TCP_TIMEOUT
  jb process_no_retransmit
  call tcp_retransmit
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset timers
  mov word ptr cs:[di], 0 ;reset timeout timer
  mov di, offset tcp_retry
  inc byte ptr cs:[bx+di] ;increment retry counter
  process_no_retransmit:
  mov di, offset tcp_retry
  cmp byte ptr cs:[bx+di], TCP_RETRIES
  jne process_tick_loop_return
  call tcp_reset
  jmp process_tick_loop_return

process_fin_wait:
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset timers
  cmp word ptr cs:[di], TCP_WAIT
  jb process_tick_loop_return
  ;call tcp_abort
  ;call tcp_send_rst
  call tcp_reset
  mov cl, TCP_STATE_CLOSED
  call tcp_change_state
  ;mov si, offset strtimeoutfin
  ;call printmsg
  jmp process_tick_loop_return

process_syn_wait:
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset timers
  cmp word ptr cs:[di], TCP_WAIT
  jb process_tick_loop_return2
  ;call tcp_abort
  call tcp_reset
  ;call tcp_send_rst
  mov cl, TCP_STATE_CLOSED
  call tcp_change_state
  ;mov si, offset strtimeoutsyn
  ;call printmsg
  jmp process_tick_loop_return

process_tick_loop_return2:
  jmp process_tick_loop_return

process_arp_wait:
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset timers
  cmp word ptr cs:[di], TCP_ARP_TIMEOUT
  jb arp_not_timeout
  jmp process_tick_loop_return
  arp_not_timeout:
  mov di, cs:[tcp_cur_handle]
  add di, offset tcp_retry
  cmp byte ptr cs:[di], 0 ;TCP_ARP_RETRIES
  jz arp_timeout_reset
  dec byte ptr cs:[di]
  call tcp_initiate_arp
  jmp process_tick_loop_return

  arp_timeout_reset:
  call tcp_reset
  call tcp_send_rst
  mov cl, TCP_STATE_CLOSED
  call tcp_change_state
  ;mov si, offset strtimeoutarp
  ;call printmsg
  jmp process_tick_loop_return

tcp_reset_timer_and_retry:
  pushall
  mov di, cs:[tcp_cur_handle]
  shl di, 1
  add di, offset timers
  mov word ptr cs:[di], 0 ;reset timeout timer
  mov bx, cs:[tcp_cur_handle]
  mov di, offset tcp_retry
  mov byte ptr cs:[bx+di], 0 ;increment retry counter
  popall
  ret

strtcpget db '[DEBUG] Got TCP packet!', 13, 10, 0
strsynget db '[DEBUG] Got SYN TCP packet!', 13, 10, 0
strsynack db '[DEBUG] Got SYN+ACK packet!', 13, 10, 0
strack    db '[DEBUG] Got ACK packet!', 13, 10, 0
strrst    db '[DEBUG] Got RST packet!', 13, 10, 0
strmatchid db '[DEBUG] Matched handle #', 0 ;13, 10, 0
strnewhand db '[DEBUG] New handle assigned!', 13, 10, 0
strclosed  db '[DEBUG] Closed and reset handle #', 0
strms db 'MILESTONE', 13, 10, 0
strwangs db 'WANGS: ', 0
strchange db '[DEBUG] State changed to #', 0
strsamatch db '[DEBUG] SYN value matches up!', 13, 10, 0
strtcpdatalen db '[DEBUG] TCP packet data length: ', 0
strsend db 'GET / HTTP/1.1', 13, 10, 'Host: www.google.com', 13, 10, 13, 10
strarpdone db 'MAC resolved:', 13, 10, 0
strsentsyn db '[DEBUG] Sent SYN TCP packet!', 13, 10, 0
strlisten db '[DEBUG] Connection request on listener port!', 13, 10, 0
strretrans db '[DEBUG] TCP packet was re-transmitted!', 13, 10, 0
strtimeoutfin db '[DEBUG] TCP socket was reset due to timeout! [FIN state]', 13, 10, 0
strtimeoutsyn db '[DEBUG] TCP socket was reset due to timeout! [SYN state]', 13, 10, 0
strtimeoutarp db '[DEBUG] TCP socket was reset due to timeout! [ARP wait]', 13, 10, 0
